# ../respawnwave/respawnwave.py

# =============================================================================
# >> IMPORTS
# =============================================================================
# Python Imports
#    Threading
from threading import Thread, Lock
#    Time
import time

# Source.Python Imports
from Source import Player
from Source import Engine
from Source import Cvar
#   Core
from core import GameEngine, ServerVar
#   Events
from events import Event
#   Messages
from messages import Chat

# =============================================================================
# >> CONSTANTS
# =============================================================================
MESSAGE_PREFIX = "\x01[\x04RespawnWave\x01]"

# =============================================================================
# >> SERVER VARIABLES
# =============================================================================
respawn_interval = ServerVar("rw_spawn_interval", \
    10.0, \
    Cvar.FCVAR_NOTIFY, \
    "The interval between each respawn wave, in seconds.")

min_respawn_wait = ServerVar("rw_min_spawn_wait", \
    3.0, \
    Cvar.FCVAR_NOTIFY, \
    "The minimum interval for a player's respawn, in seconds.")

health_reduction_scale = ServerVar("rw_health_reduction_scale", \
    0.75, \
    Cvar.FCVAR_NOTIFY, \
    "The amount to scale a player's health by on each respawn.")                            

# =============================================================================
# >> RESPAWN WAVE THREAD CLASS
# =============================================================================
# Global thread instance
respawn_wave_thread = None

# RespawnWaveThread Class
class RespawnWaveThread(Thread):
    """
    A threading class which manages the respawning of players in waves, a
    little bit like how TF2 works in some gamemodes. 
    """    
    # Thread states
    STATE_ACTIVE = 0
    STATE_INACTIVE = 1
    STATE_CLOSED = 2
    
    def __init__(self):
        """
        Initializes the respawn wave thread, prior to starting. 
        """
        self.lock = Lock()
        self.entities = []
        self.wait_entities = []
        self.entity_healths = {} 
        self.state = RespawnWaveThread.STATE_INACTIVE
        self.next_respawn = 0.0
        Thread.__init__(self)        
    
    def append(self, entity):
        """
        Adds a [player] entity to be respawned by the respawn wave thread.
        Will not add entities if the respawner is not in STATE_ACTIVE.
        """
        # Don't add entities if the respawner is not active
        if self.state != RespawnWaveThread.STATE_ACTIVE:
            return
            
        # Start our critical section
        self.lock.acquire()
        
        # Get the remaining wait time
        wait_time = self.next_respawn - time.time()
        
        # If the wait is less than the minimum, then add to the wait list,
        # otherwise, add to the next respawn list
        if wait_time < float(min_respawn_wait):
            self.wait_entities.append(entity)
            wait_time = wait_time + float(respawn_interval)
        else:
            self.entities.append(entity)
            
        # Set the next health for the entity
        entity_index = Engine.IndexOfEdict(entity)
        if entity_index not in self.entity_healths:
            self.entity_healths[entity_index] = 100
        self.entity_healths[entity_index] *= float(health_reduction_scale)
        
        # Send a message to the [player] entity
        Chat(entity_index, \
             "%s Respawn in \x04%.1f \x01seconds..." % \
             (MESSAGE_PREFIX, wait_time))
             
        # End our critical section
        self.lock.release()
    
    def changeState(self, state):
        """
        Changes the state of the respawn wave thread.
        """
        # Start our critical section
        self.lock.acquire()
        
        # Set the state
        self.state = state
        
        # Empty the current values
        self.entities = []
        self.wait_entities = []
        self.entity_healths = {}
        
        # End our critical section
        self.lock.release()
            
    def respawnPlayer(self, entity, health):
        """
        Respawns the [player] entity, with a desired health value.
        """
        # Set the initial entity properties
        entity.SetPropInt('CCSPlayer.m_iPlayerState', 0)
        entity.SetPropInt('CBasePlayer.m_lifeState', 512)
        
        # Dispatch a spawn message for the entity
        entity.DispatchSpawn()
        
        # Set the entity's health valud
        entity.SetPropInt('CBasePlayer.m_iHealth', int(health))
        
        # Send a message to the [player] entity
        entity_index = Engine.IndexOfEdict(entity)
        health_reduction_percent = (1 - float(health_reduction_scale)) * 100
        if health_reduction_percent > 0:    
            Chat(entity_index, \
                 "%s Health has been \x07reduced by %.1f%%\x01." % \
                 (MESSAGE_PREFIX, health_reduction_percent))
        elif health_reduction_percent < 0:
            Chat(entity_index, \
                 "%s Health has been \x04increased by %.1f%%\x01." % \
                 (MESSAGE_PREFIX, health_reduction_percent * -1))
                    
                
    def run(self):
        """
        The main execution method. Invoked internally by the threading module.
        Will continue to execute until the state of the thread is set to
        STATE_CLOSED.
        """
        # Change the state to STATE_ACTIVE
        self.changeState(RespawnWaveThread.STATE_ACTIVE)

        try:                
            # Loop until the state is STATE_CLOSED        
            while self.state != RespawnWaveThread.STATE_CLOSED:                        
                # Set the next respawn time
                self.next_respawn = time.time() + float(respawn_interval)
    
                # Sleep for the wave time interval
                time.sleep(float(respawn_interval))
                
                # If the state is STATE_CLOSED, exit the loop now
                if self.state == RespawnWaveThread.STATE_CLOSED:
                    break
                    
                # Start our critical section
                self.lock.acquire()
                
                # Respawn all the [player] entities in the respawn list
                for entity in self.entities:
                    entity_index = Engine.IndexOfEdict(entity)
                    health = self.entity_healths[entity_index] 
                    self.respawnPlayer(entity, health)
                    
                # Move the waiting entity list to the respawn list
                self.entities = self.wait_entities
                
                # Empty the waiting entity list
                self.wait_entities = []
                
                # End our critical section
                self.lock.release()
        except:
            try:
                self.lock.release()
            except:
                pass

# =============================================================================
# >> ADDON LOAD/UNLOAD
# =============================================================================
def load():
    """
    Called upon addon load.
    """
    global respawn_wave_thread
    # Create a RespawnWaveThread instance, and start it
    respawn_wave_thread = RespawnWaveThread()
    respawn_wave_thread.start()
    
def unload():
    """
    Called upon addon unload.
    """
    global respawn_wave_thread
    # Tell the RespawnWaveThread instance to close when it can
    respawn_wave_thread.changeState(RespawnWaveThread.STATE_CLOSED)

# =============================================================================
# >> EVENT METHODS
# =============================================================================
@Event
def player_death(GameEvent):
    """
    Called upon a player death.
    """
    global respawn_wave_thread
    # Get the victim's entity instance
    victim = GameEvent.GetInt("userid")
    victim_entity = Player.EdictOfUserid(victim)
    
    # Append the victim to the respawn wave thread    
    respawn_wave_thread.append(victim_entity)
    
@Event
def round_start(GameEvent):
    """
    Called upon a round start.
    """
    global respawn_wave_thread
    # Change the state to STATE_ACTIVE
    respawn_wave_thread.changeState(RespawnWaveThread.STATE_ACTIVE)
    Chat([], "%s Respawn Waves is enabled." % MESSAGE_PREFIX)
    
@Event
def round_end(GameEvent):
    """
    Called upon a round end.
    """
    global respawn_wave_thread
    # Change the state to STATE_INACTIVE and join
    respawn_wave_thread.changeState(RespawnWaveThread.STATE_INACTIVE)   
    
